# -*- coding: utf-8 -*-

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
---------------> please fill out this section <---------------

Your Name : Alwin Zulehner

Your E-Mail : alwin.zulehner@jku.at

Description of the algorithm :

- How does the algorithm work?
- Did you use any previously published schemes? Cite relevant papers.

In order to efficiently tackle the problem, our algorithm employs a combination of three steps: a pre-process step (aiming for reducing the complexity beforehand), a powerful but still easy to implement search method (solving the actual mapping problem), and eventually post-mapping optimization (exploiting further optimization potential after the mapping). The three steps work as follows:

First, we perform a pre-processing step to reduce the complexity (i.e. the number of gates) of the circuit to be mapped. To this end, we topologically sort the circuit and group all gates that act e.g. on qubits q0 and q1. This includes single qubit gates on q0 or q1 as well as CNOTs with control q0 and target q1 (or vice versa). We do this in a greedy fashion - until observing a CNOT with control or target q0 (q1) which acts on a qubit different from q1 (q0).

Afterwards, the actual mapping problem is considered. To this end, an A* search algorithm is utilized. Note that A* has already successfully been applied for similar mapping tasks, e.g. in our paper entitled "Efficient Mapping of Quantum Circuits to the IBM QX Architecture" (available at https://arxiv.org/abs/1712.04722). The mapping algorithm can be formulated as follows: We add all groups of gates to the compiled circuit, where the corresponding logical qubits are mapped to physical ones that are connected in the coupling map. Afterwards, we determine the set of groups that could be added next /(according to their topological order). Since the current mapping does not allow to apply those gates, we use an A* search algorithm to determine the cheapest sequence of SWAP gates (regarding a certain heuristic) such that least one of the groups in this set can be applied. These two steps are repeated until all gates are added to the compiled circuit. Moreover, initial mapping is determined throughout this procedure.

Finally, we apply post-mapping optimizations to the groups gates (including the added SWAP gates). To this end, we build a single 4x4 unitary matrix representing the functionality of the whole group, decompose it by using the KAK-decomposition provided in qiskit, and optimize subsequent single qubit gates (note that we use fallbacks since the respective implementations in qiskit fail sometimes).

- What packages did you use in your code and for what part of the algorithm?

The algorithm requires the package Cython, since the A* algorithm is written in Cython in order to improve performance.

- How general is your approach? Does it work for arbitrary coupling layouts (qubit number)?

The algorithm is general and can be applied for an arbitrary number of qubits and arbitrary coupling layouts.

- Are there known situations when the algorithm fails?

No, there only exist some cases where the KAK decomposition provided by qiskit failes. However, for these cases fallbacks are provided such that the algorithm still yields a correct result. 

---------------> please fill out this section <---------------
"""

# Include any Python modules needed for your implementation here

# The following class is the input and output circuit representation for a
# QISKit compiler
from qiskit.dagcircuit import DAGCircuit

import pyximportcpp; pyximportcpp.install()
import a_star_mapper_challenge
import pre_processing
import post_mapping_optimization
import time

def compiler_function(dag_circuit, coupling_map=None, gate_costs=None):
    """
    Modify a DAGCircuit based on a gate cost function.

    Instructions:
        Your submission involves filling in the implementation
        of this function. The function takes as input a DAGCircuit
        object, which can be generated from a QASM file by using the
        function 'qasm_to_dag_circuit' from the included 
        'submission_evaluation.py' module. For more information
        on the DAGCircuit object see the or QISKit documentation
        (eg. 'help(DAGCircuit)').

    Args:
        dag_circuit (DAGCircuit): DAGCircuit object to be compiled.
        coupling_circuit (list): Coupling map for device topology.
                                 A coupling map of None corresponds an
                                 all-to-all connected topology.
        gate_costs (dict) : dictionary of gate names and costs.

    Returns:
        A modified DAGCircuit object that satisfies an input coupling_map
        and has as low a gate_cost as possible.
    """

    
    #####################
    # Put your code here
    ##################### 

    import copy
    from qiskit.mapper import Coupling
    from qiskit import qasm, unroll
    import networkx as nx

    if gate_costs == None:
        gate_costs = {'id': 0, 'u1': 0, 'measure': 0, 'reset': 0, 'barrier': 0, 'u2': 1, 'u3': 1, 'U': 1, 'cx': 10, 'CX': 10}

    compiled_dag = copy.deepcopy(dag_circuit)
    
    # temporary circuit to add all used gates to the available gate set
    tmp_qasm = "OPENQASM 2.0;\n" + \
                   "gate cx c,t { CX c,t; }\n" + \
                   "gate u3(theta,phi,lambda) q { U(theta,phi,lambda) q; }\n" + \
                   "gate u2(phi,lambda) q { U(pi/2,phi,lambda) q; }\n" + \
                   "gate u1(lambda) q { U(0,0,lambda) q; }\n" + \
                   "qreg q[2];\n" + \
                   "cx q[0], q[1];\n" + \
                   "u3(0.1,0.4,0.7) q[0];\n" + \
                   "u2(0.1,0.4) q[0]\n;" + \
                   "u1(0.1) q[0];\n"
    u = unroll.Unroller(qasm.Qasm(data=tmp_qasm).parse(),
                        unroll.DAGBackend(["cx", "u3", "u2", "u1"]))
    tmp_circuit = u.execute()

    # prepare empty circuit for the result
    empty_dag = DAGCircuit()

    coupling = Coupling(coupling_map)
    empty_dag.add_qreg('q', coupling.size())

    for k, v in sorted(compiled_dag.cregs.items()):
        empty_dag.add_creg(k, v)

    empty_dag.basis = compiled_dag._make_union_basis(tmp_circuit)
    empty_dag.gates = compiled_dag._make_union_gates(tmp_circuit) 

    # pre processing: group gates
    grouped_gates = pre_processing.group_gates(compiled_dag)

    # call mapper (based on an A* search) to satisfy the constraints for CNOTs given by the coupling_map 
    compiled_dag = a_star_mapper_challenge.a_star_mapper(copy.deepcopy(grouped_gates), coupling_map, coupling.size(), copy.deepcopy(empty_dag))
    grouped_gates_compiled = pre_processing.group_gates(compiled_dag)

    # estimate the cost of the mapped circuit: the number of groups as well as the cost regarding to gate_costs   
    min_groups = grouped_gates_compiled.order()
    min_cost = 0
    for op, count in compiled_dag.count_ops().items():
        min_cost += count * gate_costs[op]

    # Repeat the mapping procedure 9 times and take the result with minimum groups/cost. Each call may yield a different result, since the mapper is implemented with a certain non-determinism. In fact, in the priority queue used for implementing the A* algorithm, the entries are a pair of the priority and a pointer to an object holding th mapping infomation (as second criterion). Thus, it is uncertain which node is expanded first if two nodes have the same priority (it depends on the value of the pointer). However, this non-determinism allows to find different solution by repeatedly calling the mapper.
    for i in range(9):
        result = a_star_mapper_challenge.a_star_mapper(copy.deepcopy(grouped_gates), coupling_map, coupling.size(), copy.deepcopy(empty_dag))
        grouped_gates_result = pre_processing.group_gates(result)

        groups = grouped_gates_result.order()
        cost = 0
        for op, count in result.count_ops().items():
            cost += count * gate_costs[op]
        # take the solution with fewer groups (fewer cost if the number of groups is equal) 
        if groups < min_groups or (groups == min_groups and cost < min_cost):
            min_groups = groups
            min_cost = cost
            compiled_dag = result
            grouped_gates_compiled = grouped_gates_result

    # post-mapping optimization: build 4x4 matrix for gate groups and decompose them using KAK decomposition.
    # Moreover, subsequent single qubit gates are optimized 
    compiled_dag = post_mapping_optimization.optimize_gate_groups(grouped_gates_compiled, coupling.get_edges(), copy.deepcopy(empty_dag), gate_costs)
   
    return compiled_dag
